/*	$Id: pyrun.c,v 1.2 1999/10/07 00:22:45 leonard Exp $	*/
/* Copyright 1999, David Leonard. The University of Queensland, Australia. */

#include <stdio.h>
#include <assert.h>
#include "Python.h"
#include "symbol.h"
#include "lalr.h"
#include "run.h"

#define MAXSTACK 256

void
printsymval(f, g, sym, val)
	FILE *f;
	grammar_t *g;
	symbol_t sym;
	void *val;
{
	PyObject *repr;

	if (sym != -1)
		fprintf(f, "%s", g->g_name[sym]);
	if (val == NULL) {
		fprintf(f, "/<exception>");
	} else if ((PyObject *)val != Py_None) {
		repr = PyObject_Repr((PyObject *)val);
		if (repr == NULL)
			PyErr_Clear();
		else {
			fprintf(f, "/");
			PyObject_Print(repr, f, Py_PRINT_RAW);
			Py_DECREF(repr);
		}
	}
}

/*
 * Run the parser state machine. A stack of symbol/state pairs is
 * created and shifts and reduce-shifts are efficiently called.
 * We call the input function (ifn) to get new symbols when needed.
 */
int
run(g, ifn, istate, redfn, resultp, eofsym)
	grammar_t *g;
	symbol_t (* ifn)(void *, void **value);
	void *istate;
	void* redfn(production_t *, void **value);
	void **resultp;
	symbol_t eofsym;
{
	PyObject *expect, *info;
	symbol_t next, sym;
	void *   next_value;
	void *   rval;
	symbol_t sym_stack[MAXSTACK];
	state_t  *state_stack[MAXSTACK];
	void	 *value_stack[MAXSTACK];
	state_t  *st;
	production_t *p;
	int top;
	int retval;
	int i;
	int ateof;
	FILE* ptracef = g->g_ptracef;

	top = 0;
	ateof = 0;

	state_stack[top] = g->g_init;
	value_stack[top] = NULL;
	sym_stack[top] = -1;
	next = (*ifn)(istate, &next_value);
	if (next == -1) {
#ifndef NDEBUG
		fprintf(stderr, "exception from first token\n");
#endif
		retval = RUN_INTERNAL;
		goto error;
	}

	if (next == eofsym)
		ateof = 1;

	for (;;) {
		if (ptracef != NULL) {
			for (i = 1; i <= top; i++) {
				fprintf(ptracef, " ");
#ifndef NDEBUG
				fprintf(ptracef, "%d:", 
				    state_stack[i] ? state_stack[i]->s_id : -1);
#endif
				printsymval(ptracef, g, 
				    sym_stack[i], value_stack[i]);
			}
			fprintf(ptracef, " . ");
			if (next != eofsym)
				printsymval(ptracef, g, next, next_value);
			fprintf(ptracef, "\n");
		}
		st = state_stack[top]->s_shift[next];
		p = state_stack[top]->s_reduce[next];

		if (st != NULL) {
			/* shift input */
			top++;
			state_stack[top] = st;
			sym_stack[top] = next;
			value_stack[top] = next_value;
			if (!ateof) {
				/* only read new tokens if eof not seen */
				next = (*ifn)(istate, &next_value);
				if (next == -1) {
					/* token generator error */
					retval = RUN_INTERNAL;
					goto error;
				}
				if (next == eofsym) {
					ateof = 1;
				}
			}

		} else if (p != NULL) {
			/* REDUCE */
			top -= p->p_nrhs;

			if (ptracef != NULL) {
				fprintf(ptracef, "reduce: %s <-",
					g->g_name[p->p_lhs]);
				for (i = 0; i < p->p_nrhs; i++) {
				    fprintf(ptracef, " ");
				    printsymval(ptracef, g,
					sym_stack[top + 1 + i],
					value_stack[top + 1 + i]);
				}
				fprintf(ptracef, "\n");
			}

			if (redfn)
				rval = (*redfn)(p, &value_stack[top + 1]);
			else {
				Py_INCREF(Py_None);
				rval = Py_None;
			}

			if (rval == NULL) {
				retval = RUN_INTERNAL;
				goto error;
			}

			/* shift nt */
			st = state_stack[top]->s_shift[p->p_lhs];
			if (top == 0 && st == NULL) {
				/* shifted goal symbol */
				if (ptracef != NULL) {
					fprintf(ptracef, "accept: ");
					printsymval(ptracef, g, p->p_lhs, rval);
					fprintf(ptracef, "\n");
				}
				if (resultp)
					*resultp = rval;
				return RUN_ACCEPT;
			}
#ifndef NDEBUG
			if (st == NULL) {
				PyErr_SetString(PyExc_RuntimeError,
					"parser broken");
				retval = RUN_INTERNAL;
				goto error;
			}
#else
			assert(st);
#endif
			top++;
			state_stack[top] = st;
			value_stack[top] = rval;
			sym_stack[top] = p->p_lhs;
		} else {
			/* not in lookahead set! */
			if (ptracef != NULL)
				fprintf(ptracef, "cannot reduce, or shift %s\n",
					g->g_name[next]);

			/*
			 * POSSIBLE IDEAS:
			 * if there is an error symbol shiftable in this state,
			 * fake and 'shift' it. when in the new state, 
			 * consume symbols until something comes up in either
			 * a shift or reduce set?
			 */
			/*
			 * For now, crash and burn. Later, will need to
			 * call some error handler that can know more about
			 * the symbol (maybe it's file position?)
			 */
			/* XXXXXXXX memory leak */
			if (next == RUN_ACCEPT) {
				retval = RUN_INTERNAL;
			}

			expect = PyList_New(0);
			for (sym = 1; sym < g->g_terminals; sym++)
			    if (state_stack[top]->s_shift[sym])
				PyList_Append(expect, 
					PyString_FromString(g->g_name[sym]));
			st = state_stack[top]->s_shift[next];
			p = state_stack[top]->s_reduce[next];
			info = Py_BuildValue("OOs", expect, next_value, 
				g->g_name[next]);
#ifndef NDEBUG
			fprintf(stderr, "exception arg: ");
			PyObject_Print(info, stderr, Py_PRINT_RAW);
			fprintf(stderr, "\n");
#endif
			*resultp = info;
			retval = RUN_SYNTAX;
			goto error;
		}
	}

error:
#ifndef NDEBUG
	fprintf(stderr, "error in run()\n");
#endif
	/* Clean up python values */
	for (i = 0; i <= top; i++)
		Py_XDECREF((PyObject *)value_stack[i]);
	return retval;
}
